/*
Szemaforok és osztott memória -Nevvel ellatott szemafor
A termelő/fogyasztó problémája:
A termelő fogyasztási cikkeket termel, melyeket elhelyez egy futószalagon (véges puffer). A fogyasztó sorban leveszi a szalagról a termékeket és elfogyasztja őket. 
Forditani: gcc szemaforNevvel.cpp -pthread -lrt
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <semaphore.h>

#define SHMOBJ_PATH	"/tmp"

struct msg_s {
	int t[10];
	char name_gyVf[10];
	char name_termekek[10];
	char name_tele[10];
};

int main(int argc, char *argv[]) {
	
	sem_t * sem_gyVf; //gyarthatunk-e v fogyaszthatunk-e
	sem_t * sem_termekek; //varakozo termekek
	sem_t * sem_tele; //tele van-e

	srandom(getpid());
	int myIndex;
	myIndex = 2; //2 ha fogyaszto
	
	struct msg_s *shared_msg; /* the shared segment, and head of the messages list */
	int shmfd;	
    	int shared_seg_size = sizeof(struct msg_s);
    
	shmfd = shm_open(SHMOBJ_PATH, O_RDWR, S_IRWXU | S_IRWXG);
    	if (shmfd < 0) {
		
		myIndex = 1; //1 ha termelo
        	/* creating the shared memory object    --  shm_open()  */
    		shmfd = shm_open(SHMOBJ_PATH, O_CREAT | O_EXCL | O_RDWR, S_IRWXU | S_IRWXG);
    		if (shmfd < 0) {
        		perror("In shm_open()");
        		exit(1);
    		}
		
		/* adjusting mapped file size (make room for the whole segment to map) --  ftruncate() */
    		ftruncate(shmfd, shared_seg_size);
    	}

	/* requesting the shared segment --  mmap() */    
    	shared_msg = (struct msg_s *)mmap(NULL, shared_seg_size, PROT_READ | PROT_WRITE, MAP_SHARED, shmfd, 0);
    	if (shared_msg == NULL) {
        	perror("In mmap()");
        	exit(1);
    	}

	if(myIndex == 1){
	//termelo
		printf("Termelo elindult.\n");
		//meghatarozza a szemaforok nevet
		strcpy(shared_msg->name_gyVf, "gyVf");
		strcpy(shared_msg->name_termekek, "termekek");
		strcpy(shared_msg->name_tele, "tele");
		//letrehozza a szemaforokat a megadott nevvel
		sem_gyVf = sem_open(shared_msg->name_gyVf, O_CREAT, S_IRUSR | S_IWUSR, 1);
		if(sem_gyVf == SEM_FAILED){
			perror("In sem_open(sem_gyVf)");
        		exit(1);
		}
		sem_termekek = sem_open(shared_msg->name_termekek, O_CREAT, S_IRUSR | S_IWUSR, 0);
		if(sem_termekek == SEM_FAILED){
			perror("In sem_open(sem_termekek)");
        		exit(1);
		}
		sem_tele = sem_open(shared_msg->name_tele, O_CREAT, S_IRUSR | S_IWUSR, 3);
		if(sem_tele == SEM_FAILED){
			perror("In sem_open(sem_tele)");
        		exit(1);
		}

		//"uriti a futoszallagot"
		for(int i=0; i < 4; i++){
			shared_msg->t[i] = 0;		
		}		
	
		//felrakja a termekeket a "futoszallagra"
		int x;
		for(int i=0; i < 10; i++){
			sem_wait(sem_tele);
			sem_wait(sem_gyVf);
			sem_getvalue(sem_termekek, &x);
			shared_msg->t[x] = i;
			printf("%d. termek legyartva!\n", i);
			sem_post(sem_termekek);
			sem_post(sem_gyVf);	
			//alszik randomra, hogy valtozatosabb legyen a megoldas
			double t = random() % 30 / 10;
			sleep(t);
		}			
	}
	else{
	//fogyaszto
		printf("Fogyaszto elindult.\n");
		//megnyitja a szemaforokat
		sem_gyVf = sem_open(shared_msg->name_gyVf, 0);
		if(sem_gyVf == SEM_FAILED){
			perror("In sem_open(sem_gyVf)");
        		exit(1);
		}
		sem_termekek = sem_open(shared_msg->name_termekek, 0);
		if(sem_termekek == SEM_FAILED){
			perror("In sem_open(sem_termekek)");
        		exit(1);
		}
		sem_tele = sem_open(shared_msg->name_tele, 0);
		if(sem_tele == SEM_FAILED){
			perror("In sem_open(sem_tele)");
        		exit(1);
		}
		//leszedi a termekeket
		int x;
		for(int i=0; i < 10; i++){
			sem_wait(sem_termekek);
			sem_wait(sem_gyVf);
			printf("%d. termek elfogyasztva!\n", shared_msg->t[0]);
			for(int j = 0; j < 3; j++)
				shared_msg->t[j] = shared_msg->t[j+1];
			sem_post(sem_tele);
			sem_post(sem_gyVf);
			//alszik o is randomra, hogy minden futtatasnal kicsit maskepp mukodjon
			double t = random() % 30 / 10;
			sleep(t);
		}	
	}

	if(myIndex == 2)
	{
		//a fogyaszto zarja be a szemaforokat
		if (sem_unlink(shared_msg->name_gyVf) != 0) 
		{
        		perror("In sem_unlink(shared_msg->sem_gyVf)");
        		exit(1);
    		}
		if (sem_unlink(shared_msg->name_termekek) != 0) 
		{
        		perror("In sem_unlink(shared_msg->sem_termekek)");
        		exit(1);
    		}
		if (sem_unlink(shared_msg->name_tele) != 0) 
		{
        		perror("In sem_unlink(shared_msg->sem_tele)");
        		exit(1);
    		}
		//es torli az osztott memoriat
    		if (shm_unlink(SHMOBJ_PATH) != 0) 
		{
        		perror("In shm_unlink(SHMOBJ_PATH)");
        		exit(1);
    		}	
	}
	
	return 0;  
}
